#
# mnemosyne1_mem.py <Peter.Bienstman@UGent.be>
#

import os
import re
import sys
import time
import shutil
import cPickle
import datetime
import calendar

from mnemosyne.libmnemosyne.translator import _
from mnemosyne.libmnemosyne.utils import expand_path
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.file_formats.science_log_parser \
     import ScienceLogParser
        
re_src = re.compile(r"""src=\"(.+?)\"""", re.DOTALL | re.IGNORECASE)
re_sound = re.compile(r"""<sound src=\".+?\">""", re.DOTALL | re.IGNORECASE)

class Mnemosyne1Mem(FileFormat):
    
    description = _("Mnemosyne 1.x *.mem files")
    filename_filter = _("Mnemosyne 1.x databases") + " (*.mem)"
    import_possible = True
    export_possible = False

    def do_import(self, filename, tag_name=None, reset_learning_data=False):
        db = self.database()
        # Manage database indexes.
        db.before_mem_import()
        # The import process generates card log entries, which we will delete
        # in favour of those events that are recorded in the logs and which
        # capture the true timestamps.
        log_index = db.current_log_index()
        result = self._import_mem_file(filename, tag_name, reset_learning_data)
        if result:
            return result
        db.remove_card_log_entries_since(log_index)
        # The events that we import from the science logs obviously should not
        # be reexported to these logs (this is true for both the archived logs
        # and log.txt). So, before the import, we flush the SQL logs to the
        # science logs, and after the import we edit the partership index to
        # skip these entries.
        db.dump_to_science_log()
        self._import_logs(filename)
        db.skip_science_log()
        # Force an ADDED_CARD log entry for those cards that did not figure in
        # the txt logs, e.g. due to missing or corrupt logs.
        db.add_missing_added_card_log_entries(\
            set(item.id for item in self.items))
        # In 2.x, repetition events are used to update a card's last_rep and
        # next_rep during sync. In 1.x, there was no such information, and
        # calculating it from the logs will fail if they are incomplete.
        # Therefore, we force a card edit event for all cards.
        timestamp = int(time.time())
        for item in self.items:
            db.log_edited_card(timestamp, item.id)
        # Mananage database indexes.
        db.after_mem_import()
        # Detect inverses.
        db.link_inverse_cards()
        db.save()
            
    def _import_mem_file(self, filename, tag_name=None,
                         reset_learning_data=False):        
        self.import_dir = os.path.dirname(os.path.abspath(filename))
        w = self.main_widget()    
        # Mimick 1.x module structure.
        class MnemosyneCore(object):                          
            class StartTime:                                    
                pass                                            
            class Category:                                     
                pass                                            
            class Item:                                         
                pass
        sys.modules["mnemosyne.core"] = object()       
        sys.modules["mnemosyne.core.mnemosyne_core"] = MnemosyneCore()       
        # Load data.
        try:
            memfile = file(filename, "rb")
            header = memfile.readline()
            self.starttime, self.categories, self.items = cPickle.load(memfile)
            self.starttime = self.starttime.time
        except:
            w.show_error(_("Unable to open file."))
            return -1
        # See if the file was imported before.
        try:
            card = self.database().card(self.items[0].id,
                is_id_internal=False)
        except:
            card = None
        if card:
            w.show_error(\
                _("This file seems to have been imported before. Aborting..."))
            return -2          
        # Convert to 2.x data structures.
        w.set_progress_text(_("Importing cards..."))
        w.set_progress_range(0, len(self.items))
        w.set_progress_update_interval(len(self.items)/50)
        count = 0
        w.set_progress_value(0)
        self.map_plugin_activated = False
        self.items_by_id = {}
        for item in self.items:
            if item.id in self.items_by_id:
               item.id = "dup" + item.id 
            self.items_by_id[item.id] = item
        for item in self.items:
            count += 1
            w.set_progress_value(count)
            self._create_card_from_item(item)
        w.set_progress_value(len(self.items))
                        
    def _import_logs(self, filename):
        w = self.main_widget()
        w.set_progress_text(_("Importing history..."))
        parser = ScienceLogParser(self.database(),
            ids_to_parse=self.items_by_id)
        log_dir = os.path.join(os.path.dirname(filename), "history")
        if not os.path.exists(log_dir):
            w.close_progress()
            w.show_information(_("No history found to import."))
            return
        filenames = [os.path.join(log_dir, logname) for logname in \
            sorted(os.listdir(unicode(log_dir))) if logname.endswith(".bz2")]       
        # log.txt can also contain data we need to import, especially on the
        # initial upgrade from 1.x. 'ids_to_parse' will make sure we only pick
        # up the relevant events. (If we do the importing after having used
        # 2.x for a while, there could be duplicate load events, etc, but these
        # don't matter.)
        filenames.append(os.path.join(os.path.dirname(filename), "log.txt"))
        w.set_progress_range(0, len(filenames))
        for count, filename in enumerate(filenames):
            w.set_progress_value(count)
            try:
                parser.parse(filename)
            except:
                w.show_information(_("Ignoring unparsable file:") + " " +\
                    filename)
        w.close_progress()

    def _create_card_from_item(self, item):
        if item.cat.name == "<default>" or item.cat.name == "":
            item.cat.name = "__UNTAGGED__"
        # Don't create 'secondary' cards here, but create them together with
        # the 'main' card, except when the 'main' card has been deleted.
        if item.id.endswith(".inv") or item.id.endswith(".tr.1"):
            if not item.id.split(".", 1)[0] in self.items_by_id:
                card_type = self.card_type_with_id("1")
                fact_data = {"f": item.q, "b": item.a}
                self._preprocess_media(fact_data) 
                card = self.controller().create_new_cards(fact_data,
                    card_type, grade=-1, tag_names=[item.cat.name],
                    check_for_duplicates=False, save=False)[0]
                self._set_card_attributes(card, item)
            return
        # Map.
        if item.id + ".inv" in self.items_by_id and \
           "answerbox: overlay" in item.q and "<img " in item.q:
            item_2 = self.items_by_id[item.id + ".inv"]
            loc = item_2.a
            marked = item_2.q
            blank = ""
            for match in re_src.finditer(item.q):
                blank = "<img %s>" % match.group()
            if self.map_plugin_activated == False:
                self._activate_map_plugin()
                self.map_plugin_activated = True
            card_type = self.card_type_with_id("4")
            fact_data = {"loc": loc, "marked": marked, "blank": blank}
            self._preprocess_media(fact_data) 
            card_1, card_2 = self.controller().create_new_cards(fact_data,
                card_type, grade=-1, tag_names=[item.cat.name],
                check_for_duplicates=False, save=False)
            self._set_card_attributes(card_2, item)
            self._set_card_attributes(card_1, item_2)
        # Front-to-back.
        elif item.id + ".inv" not in self.items_by_id and \
           item.id + ".tr.1" not in self.items_by_id:
            card_type = self.card_type_with_id("1")
            fact_data = {"f": item.q, "b": item.a}
            self._preprocess_media(fact_data) 
            card = self.controller().create_new_cards(fact_data,
                card_type, grade=-1, tag_names=[item.cat.name],
                check_for_duplicates=False, save=False)[0]
            self._set_card_attributes(card, item)
        # Front-to-back and back-to-front.         
        elif item.id + ".inv" in self.items_by_id:
            card_type = self.card_type_with_id("2")
            fact_data = {"f": item.q, "b": item.a}
            self._preprocess_media(fact_data) 
            card_1, card_2 = self.controller().create_new_cards(fact_data,
                card_type, grade=-1, tag_names=[item.cat.name],
                check_for_duplicates=False, save=False)
            self._set_card_attributes(card_1, item)
            self._set_card_attributes(card_2,
                                      self.items_by_id[item.id + ".inv"])               
        # Vocabulary.
        elif item.id + ".tr.1" in self.items_by_id:
            card_type = self.card_type_with_id("3")
            try:
                p_1, m_1 = item.a.split("\n", 1)
            except:
                p_1, m_1 = "", item.a    
            fact_data = {"f": item.q, "p_1": p_1, "m_1": m_1}
            self._preprocess_media(fact_data) 
            card_1, card_2 = self.controller().create_new_cards(fact_data,
                card_type, grade=-1, tag_names=[item.cat.name],
                check_for_duplicates=False, save=False)            
            self._set_card_attributes(card_1, item)
            self._set_card_attributes(card_2,
                                      self.items_by_id[item.id + ".tr.1"])
            
    def _midnight_UTC(self, timestamp):
        date_only = datetime.date.fromtimestamp(timestamp)
        return int(calendar.timegm(date_only.timetuple()))
    
    def _set_card_attributes(self, card, item):
        # Note that we cannot give cards a new id, otherwise the log server
        # would not know it was the same card.
        self.database().change_card_id(card, item.id)
        for attr in ["id", "grade", "easiness", "acq_reps", "ret_reps",
            "lapses", "acq_reps_since_lapse", "ret_reps_since_lapse"]:
            setattr(card, attr, getattr(item, attr))    
        DAY = 24 * 60 * 60 # Seconds in a day.
        card.last_rep = \
            self._midnight_UTC(self.starttime + item.last_rep * DAY)
        card.next_rep = \
            self._midnight_UTC(self.starttime + item.next_rep * DAY)
        if item.unseen and item.grade in [0, 1]:
            card.grade = -1
            card.acq_reps = 0
            card.acq_reps_since_lapse = 0
            card.last_rep = -1
            card.next_rep = -1
        self.database().update_card(card)
        
    def _preprocess_media(self, fact_data):
        media_dir = self.database().media_dir()
        # os.path.normpath does not convert Windows separators to Unix
        # separators, so we need to make sure we internally store Unix paths.
        for key in fact_data:
            for match in re_src.finditer(fact_data[key]):
                fact_data[key] = fact_data[key].replace(match.group(),
                            match.group().replace("\\", "/"))
        # Convert sound tags to audio tags.
        for key in fact_data:
            for match in re_sound.finditer(fact_data[key]):
                print match.group(), match.group(0)
                fact_data[key] = fact_data[key].replace(match.group(),
                    match.group().replace("<sound src", "<audio src"))
        # Copy files to media directory, creating subdirectories as we go.
        for key in fact_data:
            for match in re_src.finditer(fact_data[key]):
                filename = match.group(1)
                if not os.path.exists(filename) and \
                    not os.path.exists(expand_path(filename, self.import_dir)):
                    self.main_widget().show_information(\
                        _("Missing media file") + " %s" % filename)
                    fact_data[key] = fact_data[key].replace(match.group(),
                        "src_missing=\"%s\"" % match.group(1))
                    continue
                if not os.path.isabs(filename):
                    source = expand_path(filename, self.import_dir)
                    dest = expand_path(filename, media_dir)
                    directory = os.path.dirname(dest)
                    if not os.path.exists(directory):
                        os.makedirs(directory)
                    shutil.copy(source, dest)

    def _activate_map_plugin(self):
        for plugin in self.plugins():
            component = plugin.components[0]
            if component.component_type == "card_type" and component.id == "4":
                plugin.activate()
    
